// SPDX-License-Identifier: AGPL-3.0-or-later
mod config;

use btfproto::{
    local_fs::{LocalFs, ModeAuthorizer},
    server::{new_fs_server, FsProvider},
};
use btlib::{
    config_helpers::from_envvar,
    crypto::{tpm::TpmCredStore, CredStore, Creds},
};
use btmsg::Receiver;
use config::{Config, ConfigRef, Envvars};
use std::{
    net::{IpAddr, Ipv6Addr},
    path::PathBuf,
    str::FromStr,
    sync::Arc,
};

const ENVVARS: Envvars<'static> = Envvars {
    ip_addr: "BTFSD_IPADDR",
    tabrmd: "BTFSD_TABRMD",
    tpm_state_path: "BTFSD_TPMSTATE",
    block_dir: "BTFSD_BLOCKDIR",
};

const DEFAULT_CONFIG: ConfigRef<'static> = ConfigRef {
    ip_addr: IpAddr::V6(Ipv6Addr::LOCALHOST),
    tabrmd: "bus_type=session",
    tpm_state_path: "./tpm_state",
    block_dir: "./bt",
};

fn provider<C: 'static + Send + Sync + Creds>(block_dir: PathBuf, creds: C) -> impl FsProvider {
    if block_dir.exists() {
        LocalFs::new_existing(block_dir, creds, ModeAuthorizer).unwrap()
    } else {
        std::fs::create_dir_all(&block_dir).unwrap();
        LocalFs::new_empty(block_dir, 0, creds, ModeAuthorizer).unwrap()
    }
}

fn receiver(config: Config) -> impl Receiver {
    let cred_store = TpmCredStore::from_tabrmd(&config.tabrmd, config.tpm_state_path).unwrap();
    let node_creds = cred_store.node_creds().unwrap();
    let provider = Arc::new(provider(config.block_dir, node_creds.clone()));
    new_fs_server(config.ip_addr, Arc::new(node_creds), provider).unwrap()
}

#[tokio::main]
async fn main() {
    let ip_addr = from_envvar(ENVVARS.ip_addr)
        .unwrap()
        .map(|txt| IpAddr::from_str(&txt).unwrap());
    let tabrmd = from_envvar(ENVVARS.tabrmd).unwrap();
    let tpm_state_path = from_envvar(ENVVARS.tpm_state_path)
        .unwrap()
        .map(PathBuf::from);
    let block_dir = from_envvar(ENVVARS.block_dir).unwrap().map(PathBuf::from);
    let config = Config::builder()
        .with_ip_addr(ip_addr)
        .with_tabrmd(tabrmd)
        .with_tpm_state_path(tpm_state_path)
        .with_block_dir(block_dir)
        .build();
    let receiver = receiver(config);
    receiver.complete().unwrap().await.unwrap();
}

#[cfg(test)]
mod tests {
    use super::*;

    use btfproto::{client::FsClient, msg::*};
    use btlib_tests::TpmCredStoreHarness;
    use std::future::ready;
    use tempdir::TempDir;

    struct TestCase {
        _ip_addr: IpAddr,
        _harness: TpmCredStoreHarness,
        _dir: TempDir,
    }

    impl TestCase {
        const ROOT_PASSWD: &str = "existential_threat";

        fn new() -> (Self, impl Receiver) {
            let dir = TempDir::new("btfsd").unwrap();
            let harness = TpmCredStoreHarness::new(Self::ROOT_PASSWD.to_owned()).unwrap();
            let ip_addr = IpAddr::V6(Ipv6Addr::LOCALHOST);
            let config = Config {
                ip_addr,
                tabrmd: harness.swtpm().tabrmd_config().to_owned(),
                tpm_state_path: harness.swtpm().state_path().to_owned(),
                block_dir: dir.path().join("bt"),
            };
            let receiver = receiver(config);
            (
                Self {
                    _dir: dir,
                    _harness: harness,
                    _ip_addr: ip_addr,
                },
                receiver,
            )
        }
    }

    #[allow(dead_code)]
    async fn manual_test() {
        let (_case, rx) = TestCase::new();
        rx.complete().unwrap().await.unwrap();
    }

    #[tokio::test]
    async fn create_write_read() {
        const FILENAME: &str = "file.txt";
        const EXPECTED: &[u8] = b"potato";
        let (_case, rx) = TestCase::new();
        let tx = rx.transmitter(rx.addr().clone()).await.unwrap();
        let client = FsClient::new(tx);

        let CreateReply { inode, handle, .. } = client
            .create(
                SpecInodes::RootDir.into(),
                FILENAME,
                FlagValue::ReadWrite.into(),
                0o644,
                0,
            )
            .await
            .unwrap();
        let WriteReply { written, .. } = client.write(inode, handle, 0, EXPECTED).await.unwrap();
        assert_eq!(EXPECTED.len() as u64, written);
        let msg = Read {
            inode,
            handle,
            offset: 0,
            size: EXPECTED.len() as u64,
        };
        let actual = client
            .read(msg, |reply| {
                let mut buf = Vec::with_capacity(EXPECTED.len());
                buf.extend_from_slice(reply.data);
                ready(buf)
            })
            .await
            .unwrap();

        assert_eq!(EXPECTED, &actual);
    }
}
