// SPDX-License-Identifier: AGPL-3.0-or-later
use super::DEFAULT_CONFIG;

use std::path::{Path, PathBuf};

#[derive(PartialEq, Eq, Clone)]
pub struct Config {
    pub block_dir: PathBuf,
    pub mnt_dir: PathBuf,
    pub tpm_state_file: PathBuf,
    pub tabrmd: String,
    pub mnt_options: String,
    pub threads: usize,
}

impl Config {
    pub fn builder() -> ConfigBuilder {
        ConfigBuilder::new()
    }
}

pub struct ConfigRef<'a> {
    pub block_dir: &'a str,
    pub mnt_dir: &'a str,
    pub tpm_state_file: &'a str,
    pub tabrmd: &'a str,
    pub mnt_options: &'a str,
    pub threads: usize,
}

#[derive(Default, PartialEq, Eq, Clone)]
pub struct ConfigBuilder {
    pub block_dir: Option<PathBuf>,
    pub mnt_dir: Option<PathBuf>,
    pub tpm_state_file: Option<PathBuf>,
    pub tabrmd: Option<String>,
    pub mnt_options: Option<String>,
    pub threads: Option<usize>,
}

impl ConfigBuilder {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn with_block_dir(mut self, block_dir: Option<PathBuf>) -> Self {
        self.block_dir = block_dir;
        self
    }

    pub fn with_mnt_dir(mut self, mnt_dir: Option<PathBuf>) -> Self {
        self.mnt_dir = mnt_dir;
        self
    }

    #[allow(dead_code)]
    pub fn with_tpm_state_file(mut self, mnt_dir: Option<PathBuf>) -> Self {
        self.tpm_state_file = mnt_dir;
        self
    }

    pub fn with_tabrmd(mut self, tabrmd: Option<String>) -> Self {
        self.tabrmd = tabrmd;
        self
    }

    pub fn with_mnt_options(mut self, mnt_options: Option<String>) -> Self {
        self.mnt_options = mnt_options;
        self
    }

    #[allow(dead_code)]
    pub fn with_threads(mut self, threads: Option<usize>) -> Self {
        self.threads = threads;
        self
    }

    pub fn build(self) -> Config {
        Config {
            block_dir: self
                .block_dir
                .unwrap_or_else(|| Path::new(DEFAULT_CONFIG.block_dir).to_owned()),
            mnt_dir: self
                .mnt_dir
                .unwrap_or_else(|| Path::new(DEFAULT_CONFIG.mnt_dir).to_owned()),
            tpm_state_file: self
                .tpm_state_file
                .unwrap_or_else(|| Path::new(DEFAULT_CONFIG.tpm_state_file).to_owned()),
            tabrmd: self
                .tabrmd
                .unwrap_or_else(|| DEFAULT_CONFIG.tabrmd.to_owned()),
            mnt_options: self
                .mnt_options
                .unwrap_or_else(|| DEFAULT_CONFIG.mnt_options.to_owned()),
            threads: self
                .threads
                .unwrap_or_else(|| DEFAULT_CONFIG.threads.to_owned()),
        }
    }
}

pub struct EnvVars {
    pub tabrmd: &'static str,
    pub mnt_options: &'static str,
}
